#! /usr/bin/python

import sys
import threading
import os
import time
from scapy.all import *
from optparse import OptionParser
from random import randint

"""
Author: Demetris Papapetrou and QSecure
Creation Date: one day in 2010
Release Date: 25/09/2014
Purpose: Proof of Concept (PoC) tool for revealing permissive firewall rules for trusted 3rd party networks/hosts
Dependencies: Scapy network packet generator

Comments: I know that my programming skills suck :-)

License
--------
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
"""

# Turn down the verbosity of scapy
conf.verb = 0

# Port 137/udp payload 02c00010000100000000000020434b4141414141414141414141414141414141414141414141414141414141410000210001
payload = "02c00010000100000000000020434b4141414141414141414141414141414141414141414141414141414141410000210001".decode("hex")

def set_parse_options():

	parser = OptionParser(usage="usage: %prog [options] target:port:protocol")
	parser.add_option("-a", dest="spoofed",
                  help="spoofed IP address and source port to be used to bypass ACL (mandatory). Acceptable format <IP_Address:Port>" )
	parser.add_option("-T", dest="tcp",
                  help="use TCP to gather IPIDs from the target. Available flags S/A/F/U/P/R" )
	parser.add_option("-U", dest="udp", action="store_true",
                  help="use UDP to gather IPIDs from the target" )
	parser.add_option("-I", dest="icmp",
                  help="use ICMP to gather IPIDs from the target P=Ping (default) T=Timestamp A=Address Mask" )
	parser.add_option("-P",dest="ip", action="store_true",
                  help="use IP Protocol to gather IPIDs from the target" )
	parser.add_option("-t",dest="tcp_port",
                  help="TCP port to use when gathering IPIDs from the target" )
	parser.add_option("-u",dest="udp_port",
                  help="UDP port to use when gathering IPIDs from the target" )
	parser.add_option("-p",dest="ip_proto",
                  help="IP protocol number to use when gathering IPIDs from the target")
	parser.add_option("--ttl", dest="ttl",
                  help="how many hops the packet will traverse")
	parser.add_option("-c", dest="packet_count",
                  help="number of IPID probe packets to send to the target")
	parser.add_option("-b", dest="burst",
                  help="number of spoofed packets to send to the target as a quick burst")
	parser.add_option("--fast", dest="fast", action="store_true",
                  help="send IPID probes to target faster than 1pps" )
	print ""
	print "Examples:"
	print "---------"
	print "Target details can be written as 192.168.1.1:22,23,80,443:T"
	print "Spoofed address details can be written as 10.10.10.1-254:53"
	print ""
	

	parser.set_defaults(tcp = None,
		udp = False,
		icmp = None,
		ip = False,
		ip_proto = 16,
		packet_count = 2,
		ttl = 128,
		burst = 1,
		fast = False)

	return parser

class ScanTCP ( threading.Thread ):

   # Override Thread's __init__ method to accept the parameters needed:
   def __init__ ( self, spoofed_ip_list, target_host, spoofed_port, target_port, burst ):

      self.spoofed_ip_list = spoofed_ip_list
      self.target_host = target_host
      self.spoofed_port = spoofed_port
      self.target_port = target_port
      self.burst = burst
      threading.Thread.__init__ ( self )

   def run ( self ):
	for spoofed_ip in spoofed_ip_list:
		send([IP(src=spoofed_ip,dst=target_host,id=RandShort())/TCP(sport=spoofed_port,dport=target_port,flags="S")] * burst)
	

class ScanUDP ( threading.Thread ):

   # Override Thread's __init__ method to accept the parameters needed:
   def __init__ ( self, spoofed_ip_list, target_host, spoofed_port, target_port, burst ):

      self.spoofed_ip_list = spoofed_ip_list
      self.target_host = target_host
      self.spoofed_port = spoofed_port
      self.target_port = target_port
      self.burst = burst
      threading.Thread.__init__ ( self )

   def run ( self ):
	for spoofed_ip in spoofed_ip_list:
		send([IP(src=spoofed_ip,dst=target_host,id=RandShort())/UDP(sport=spoofed_port,dport=target_port)/payload] * burst)
	


if __name__ == "__main__":

	parser = set_parse_options()
	(option, args) = parser.parse_args()
	
	#### Section for checking the validity of program parameters ####
	if len(args) != 1:
		parser.print_help()
        	parser.error("incorrect number of arguments")
		sys.exit(2)
	
	try:
		if option.spoofed: 
			spoofed = str(option.spoofed)
	except:
		parser.print_help()
		parser.error("spoofed host must be an IP address")
	
	try:
		if option.packet_count: 
			count = int(option.packet_count)
	except:
		parser.print_help()
		parser.error("Packet count must be an integer")
		
	try:
		if option.burst: 
			burst = int(option.burst)
	except:
		parser.print_help()
		parser.error("Packet burst must be an integer > 0")
	
	try:
		if option.tcp: 
			tcp = str(option.tcp)
	except:
		parser.print_help()
		parser.error("TCP probes must contain one or more of the flags S/A/F/P/U/R")
	
	try:
		if option.icmp: 
			icmp = str(option.icmp)
	except:
		parser.print_help()
		parser.error("ICMP probes must contain one of the following option P/T/A")
	
	try:
		if option.tcp_port: 
			tcp_port = int(option.tcp_port)
	except:
		parser.print_help()
		parser.error("TCP port must be an integer")
		
	try:
		if option.udp_port: 
			udp_port = int(option.udp_port)
	except:
		parser.print_help()
		parser.error("UDP port must be an integer")
	
	try:
		if option.ip_proto: 
			ip_proto = int(option.ip_proto)
	except:
		parser.print_help()
		parser.error("IP proto must be an integer < 255")
	
	try:
		if option.ttl: 
			ip_ttl = int(option.ttl)
	except:
		parser.print_help()
		parser.error("TTL value must be an integer < 255")
	
	
	#Check packet count 
	if count < 2:
		parser.print_help()
		parser.error("Error: Packet count must be at least 2")
		sys.exit(2)
	
	#Check burst number 
	if burst < 1:
		parser.print_help()
		parser.error("Error: Packet burst must be an integer > 0")
		sys.exit(2)
		
	#Check TCP port
	if option.tcp_port:
		tcp_port = int(option.tcp_port)
	else:
		tcp_port = 80
	tcp_sport = tcp_port
	
	#Check UDP port
	if option.udp_port:
		udp_port = int(option.udp_port)
	else:
		udp_port = 80
	udp_sport = udp_port
	
	#### PARAMETER CHECK ####
	if count >= 256:
		print "count must be an integer < 256"
		sys.exit(2)
	
	if tcp_port >= 65536 or udp_port >= 65536:
		print "UDP/TCP ports must be an integer < 65536"
		sys.exit(2)
	
	if ip_proto >= 255:
		print "IP proto must be an integer < 255"
		sys.exit(2)
	
	if ip_ttl >= 255:
		print "TTL value must be an integer < 255"
		sys.exit(2)

	#Check if only the allowed flags can be set for TCP scanning
	if option.tcp != None and tcp.find("S") ==-1 and tcp.find("A") ==-1 and tcp.find("F") ==-1 and tcp.find("P") ==-1 and tcp.find("U") ==-1 and tcp.find("R") ==-1:
		parser.print_help()
		parser.error("TCP probes must contain one or more of the flags S/A/F/P/U/R")
		sys.exit(2)
	
	#Check if only the allowed options can be set for ICMP scanning
	if option.icmp != None and icmp.find("P") ==-1 and icmp.find("T") ==-1 and icmp.find("A") ==-1:
		parser.print_help()
		parser.error("ICMP probes must contain one of the following option P/T/A")
		sys.exit(2)	
	
	#If no scan mode is defined default to ICMP	
	if option.tcp == None and option.udp == False and option.icmp == None and option.ip == False:
		option.icmp = "P"
		icmp = "P"
	
	#Separate Target's IP from Port number and Protocol
	target_port_list = list()
	if args[0]:
		target = str(args[0])
		if target.find(":") != -1:
			target_list = target.split(":",2)
			if len(target_list[:]) == 3:
				target_ip = str(target_list[0])
				target_port_string = str(target_list[1])
				target_protocol = str(target_list[2])
			elif len(target_list[:]) == 2:
				target_ip = str(target_list[0])
				target_port_string = str(target_list[1])
				target_protocol = "T"
			else:
				target_ip = target
				target_port_string = 80
				target_protocol = "T"
		
			if target_ip.find(":") != -1 or target_port_string.find(":") != -1 or target_protocol.find(":") != -1:
				print "Error: Target IP, Port and Protocol must be separated by a colon (:)"
				sys.exit(2)
			else:
				if target_port_string.find(",") != -1:
					target_port_list = target_port_string.split(",")
					for test_port in target_port_list[:]:
						if int(test_port) > 65535:
							print "Error: Target port must be an integer < 65536"
							sys.exit(2)
				else:
					target_port_int = int(target_port_string)
					if target_port_int > 65535:
						print "Error: Target port must be an integer < 65536"
						sys.exit(2)
					else:
						target_port_list.append(target_port_int)
				
				if len(target_protocol) != 1 or (target_protocol != "T" and target_protocol != "U"):
					print "Error: Target protocol must be either T or U."
					sys.exit(2)
		else:
			target_ip = target
			target_port_list.append(80)
			target_protocol = "T"
		
	#Check that spoofed IP is defined and separate IP from Port number
	spoofed_ip_list = list()
	if option.spoofed == None:	
		print "Error: A spoofed host IP must be defined. use -a <IP_Address:Port> "
		sys.exit(2)
	else:
		if spoofed.find(":") != -1:
			spoofed_list = spoofed.split(":",1)
			spoofed_ip_string = str(spoofed_list[0])
			spoofed_port = str(spoofed_list[1])
			if spoofed_ip_string.find(":") != -1 or spoofed_port.find(":") != -1:
				print "Error: Spoofed IP and port must be separated by a single colon (:)"
				sys.exit(2)
			else:
				spoofed_port = int(spoofed_port)
				if spoofed_port > 65535:
					print "Error: Spoofed source port must be an integer < 65536"
					sys.ext(2)
				#Create a List with the range of IPs
				if spoofed_ip_string.find("-") != -1:
					two_spoofed_ips_list = spoofed_ip_string.split("-")
					starting_spoofed_ip = str(two_spoofed_ips_list[0])
					spoofed_ip_host_end = int(two_spoofed_ips_list[1])
					if starting_spoofed_ip.find(".") != -1:
						starting_spoofed_ip_list = starting_spoofed_ip.rsplit(".",1)
						spoofed_ip_net = starting_spoofed_ip_list[0]
						spoofed_ip_host_start = int(starting_spoofed_ip_list[1])
						for host in range(spoofed_ip_host_start, spoofed_ip_host_end + 1, 1):
							reconstructed_ip = spoofed_ip_net + "." + str(host)
							spoofed_ip_list.append(reconstructed_ip)
					else:
						print "Error: No valid spoofed IP address found"
						sys.exit(2)
				else:
					spoofed_ip_list.append(spoofed_ip_string)
					
		#Create a List with the range of IPs
		elif spoofed.find(":") == -1 and spoofed.find("-") != -1:
			spoofed_ip_string = spoofed
			two_spoofed_ips_list = spoofed_ip_string.split("-")
			starting_spoofed_ip = str(two_spoofed_ips_list[0])
			spoofed_ip_host_end = int(two_spoofed_ips_list[1])
			if starting_spoofed_ip.find(".") != -1:
				starting_spoofed_ip_list = starting_spoofed_ip.rsplit(".",1)
				spoofed_ip_net = starting_spoofed_ip_list[0]
				spoofed_ip_host_start = int(starting_spoofed_ip_list[1])
				for host in range(spoofed_ip_host_start, spoofed_ip_host_end + 1, 1):
					reconstracted_ip = spoofed_ip_net + "." + str(host)
					spoofed_ip_list.append(reconstracted_ip)
			else:
				print "Error: No valid spoofed IP address found"
				sys.ext(2)
			spoofed_port = 80
			print "You haven't specified a spoofed source port. Using port 80."	
		else:
			spoofed_ip_list.append(spoofed)
			#print "spoofed ip: " + spoofed
			spoofed_port = 80
			print "You haven't specified a spoofed source port. Using port 80."	
		
	print "Scan process initiated..."
	##########  TCP  ##########
	try:
		#TCP probes
		if option.tcp != None:
			for target_port in target_port_list[:]:
				previousID = 0
				for i in range(count):
					if option.fast != True:
						time.sleep(1)
					ans,unans = sr(IP(dst=target_ip,id=RandShort(),ttl=ip_ttl)/TCP(sport=RandShort(),dport=tcp_port,flags=tcp),timeout=2)
					for s,r in ans:
						if r.haslayer(IP):
							if previousID == 0: previousID = r[IP].id
							if r.haslayer(TCP):
								if s[TCP].dport == r[TCP].sport:
									print "TCP:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
									target_host = r[IP].src
							elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
								print "TCP:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								target_host = r[IP].src
							elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 3:
								print "TCP:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Unreachable)"
								target_host = r[IP].src
							elif r.haslayer(ICMP):
								print "TCP:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								target_host = r[IP].src
							previousID = r[IP].id
					for s in unans:
						print "TCP: No Response"
						sys.exit(1)
					if (i == (count / 2) - 1 and i < 10) or (i == 10 and (count / 2) - 1 >= 10):
						print "TCP: sending " + str(burst) + " spoofed packet(s)... to port " + str(target_port)
						if target_protocol == "T": ScanTCP( spoofed_ip_list, target_host, spoofed_port, target_port, burst ).start()
						if target_protocol == "U": ScanUDP( spoofed_ip_list, target_host, spoofed_port, target_port, burst ).start()
				print "----------------------------------------------------------------------"	
			
		##########  UDP  ##########
		if option.udp == True:  
			for target_port in target_port_list[:]:
				previousID = 0
				for i in range(count):
					if option.fast != True:
						time.sleep(1)
					ans,unans = sr(IP(dst=target_ip,id=RandShort(),ttl=ip_ttl)/UDP(sport=RandShort(),dport=udp_port),timeout=2)
					for s,r in ans:
						if r.haslayer(IP):
							if previousID == 0: previousID = r[IP].id
							if r.haslayer(ICMP) and r.getlayer(ICMP).type == 3:
								print "UDP:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Unreachable)"
								target_host = r[IP].src
							elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
								print "UDP:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								target_host = r[IP].src
							elif r.haslayer(ICMP):
								print "UDP:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								target_host = r[IP].src
							previousID = r[IP].id	
					for s in unans:
						print "UDP: No Response. Try to find a closed port."
						sys.exit(1)
					if (i == (count / 2) - 1 and i < 10) or (i == 10 and (count / 2) - 1 >= 10):
						print "UDP: sending " + str(burst) + " spoofed packet(s)..." + str(target_port)
						if target_protocol == "T": ScanTCP( spoofed_ip_list, target_host, spoofed_port, target_port, burst ).start()
						if target_protocol == "U": ScanUDP( spoofed_ip_list, target_host, spoofed_port, target_port, burst ).start()
			print "----------------------------------------------------------------------"	
		
		##########  ICMP  ##########
		if option.icmp != None:
			for target_port in target_port_list[:]:
				previousID = 0
				for i in range(count):
					if option.fast != True:
						time.sleep(1)
					#ICMP ping scan
					if icmp.find("P") >= 0 and icmp.find("T") == -1 and icmp.find("A") == -1:
						ans,unans = sr(IP(dst=target_ip,id=RandShort(),ttl=ip_ttl)/ICMP(type=8,code=0),timeout=2)
					#ICMP timestamp scan
					elif icmp.find("P") == -1 and icmp.find("T") >= 0 and icmp.find("A") == -1:
						ans,unans = sr(IP(dst=target_ip,id=RandShort(),ttl=ip_ttl)/ICMP(type=13,code=0),timeout=2)
					#ICMP address mask scan
					elif icmp.find("P") == -1 and icmp.find("T") == -1 and icmp.find("A") >= 0:
						ans,unans = sr(IP(dst=target_ip,id=RandShort(),ttl=ip_ttl)/ICMP(type=17,code=0),timeout=2)
					else:
						print "Error: You can only use one of the ICMP options"
						sys.exit(2)
						
					for s,r in ans:
						if r.haslayer(IP):
							if previousID == 0: previousID = r[IP].id
							if r.haslayer(ICMP) and r.getlayer(ICMP).type == 0:
								print "ICMP Ping:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								target_host = r[IP].src
							elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 14:
								print "ICMP Timestamp:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								target_host = r[IP].src
							elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 18:
								print "ICMP Address Mask:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID)
								target_host = r[IP].src
							elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
								print "ICMP:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								target_host = r[IP].src
							elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 3:
								print "ICMP:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Unreachable)"
								target_host = r[IP].src
							elif r.haslayer(ICMP):
								print "ICMP:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								target_host = r[IP].src
							previousID = r[IP].id
					for s in unans:
						print "ICMP: No Response"
						sys.exit(1)
					if (i == (count / 2) - 1 and i < 10) or (i == 10 and (count / 2) - 1 >= 10):
						print "ICMP: sending " + str(burst) + " spoofed packet(s)..." + str(target_port)
						if target_protocol == "T": ScanTCP( spoofed_ip_list, target_host, spoofed_port, target_port, burst ).start()
						if target_protocol == "U": ScanUDP( spoofed_ip_list, target_host, spoofed_port, target_port, burst ).start()
						
			print "----------------------------------------------------------------------"		
					
		##########  IP  ##########
		if option.ip == True:
			for target_port in target_port_list[:]:
				previousID = 0
				for i in range(count):
					if option.fast != True:
						time.sleep(1)
					ans,unans = sr(IP(dst=target_ip,id=RandShort(),proto=ip_proto,ttl=ip_ttl)/"PIZZA",timeout=2)
					for s,r in ans:
						if r.haslayer(IP):
							if previousID == 0: previousID = r[IP].id
							if r.haslayer(ICMP) and r.getlayer(ICMP).type == 11 and r.getlayer(ICMP).code == 0:
								print "IP Protocol:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_TTL_Exceeded)"
								target_host = r[IP].src
							elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 3 and r.getlayer(ICMP).code == 2:
								print "IP Protocol:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Proto-Unreach)"
								target_host = r[IP].src
							elif r.haslayer(ICMP) and r.getlayer(ICMP).type == 3:
								print "IP Protocol:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Unreachable)"
								target_host = r[IP].src
							elif r.haslayer(ICMP):
								print "IP Protocol:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	(ICMP_Other)"
								target_host = r[IP].src
							else:
								print "IP Protocol:  ip=" + r[IP].src + " 	id=" + str(r[IP].id) + " 	inc=" + str(r[IP].id - previousID) + " 	proto=" + str(r[IP].proto)
								target_host = r[IP].src
							previousID = r[IP].id	
								
					for s in unans:
						print "IP Protocol: No Response"
						sys.exit(1)
					if (i == (count / 2) - 1 and i < 10) or (i == 10 and (count / 2) - 1 >= 10):
						print "IP Protocol: sending " + str(burst) + " spoofed packet(s)..." + str(target_port)
						if target_protocol == "T": ScanTCP( spoofed_ip_list, target_host, spoofed_port, target_port, burst ).start()
						if target_protocol == "U": ScanUDP( spoofed_ip_list, target_host, spoofed_port, target_port, burst ).start()
			print "----------------------------------------------------------------------"			
	except KeyboardInterrupt:
		print "Operation interrupted by the user..."
		sys.exit(1)
